eco_gage_set = {'A', 'AURO', 'BEAR', 'BUCK', 'BURN',
                'CAST', 'CLA2', 'CLAY',
                'CLIN', 'DURH',
                'FLET',
                'FRYI',
                'GARM',
                'GATH',
                'GBAX',
                'GBRA',
                'GBRE',
                'GBYR',
                'GCAM',
                'GCHA',
                'GCHS',
                'GCML',
                'GCOH',
                'GCOO',
                'GDAL',
                'GDAW',
                'GDIN',
                'GFTB',
                'GLOU',
                'GMCI',
                'GMCR',
                'GMET',
                'GMID',
                'GNEW',
                'GOCO',
                'GOKE',
                'GOLD',
                'GPLA',
                'GSTA',
                'GSTE',
                'GTAL',
                'GTOC',
                'GWAS',
                'GWAY',
                'HAML',
                'JACK',
                'JEFF',
                'K0A9',
                'K1A5',
                'K3J7',
                'K6A1',
                'K6A2',
                'KACJ',
                'KAFP',
                'KAIK',
                'KARW',
                'KASJ',
                'KAYS',
                'KBBP',
                'KBGE',
                'KBGF',
                'KBIJ',
                'KBNL',
                'KBQK',
                'KCCO',
                'KCDN',
                'KCKF',
                'KCKI',
                'KCNI',
                'KCPC',
                'KCQW',
                'KCTJ',
                'KCTZ',
                'KCWV',
                'KDBN',
                'KDCM',
                'KDNN',
                'KDPL',
                'KDQH',
                'KDYB',
                'KDYR',
                'KEDE',
                'KEHO',
                'KEOE',
                'KETC',
                'KEXX',
                'KEYF',
                'KEZM',
                'KFBG',
                'KFDW',
                'KFFA',
                'KFQD',
                'KGEV',
                'KGGE',
                'KGSB',
                'KGWW',
                'KGYH',
                'KHBI',
                'KHFF',
                'KHNZ',
                'KHOE',
                'KHQU',
                'KHRJ',
                'KHVS',
                'KHXD',
                'KHYW',
                'KIIY',
                'KINS',
                'KIPJ',
                'KISO',
                'KIXA',
                'KJES',
                'KJNX',
                'KJQF',
                'KJYL',
                'KJZI',
                'KLGC',
                'KLHW',
                'KLHZ',
                'KLKR',
                'KLQK',
                'KLRO',
                'KLSF',
                'KLUX',
                'KLZU',
                'KMAO',
                'KMGE',
                'KMGR',
                'KMKS',
                'KMLJ',
                'KMMI',
                'KMMT',
                'KMNI',
                'KMQI',
                'KMQY',
                'KMRN',
                'KMWK',
                'KMYR',
                'KNBC',
                'KNBT',
                'KNCA',
                'KNJM',
                'KNKT',
                'KNQA',
                'KOAJ',
                'KOCW',
                'KONX',
                'KOPN',
                'KPGV',
                'KPOB',
                'KRBW',
                'KRCZ',
                'KRHP',
                'KRUQ',
                'KRYY',
                'KSBO',
                'KSCR',
                'KSIF',
                'KSMS',
                'KSOP',
                'KSPA',
                'KSSC',
                'KSUT',
                'KSVH',
                'KSVN',
                'KTBR',
                'KTDF',
                'KTHA',
                'KTMA',
                'KTNB',
                'KTTA',
                'KTVI',
                'KUCY',
                'KUDG',
                'KUKF',
                'KVAD',
                'KVDI',
                'KVUJ',
                'KWDR',
                'KWRB',
                'LAKE',
                'LAUR',
                'LEWS',
                'LILE',
                'MITC',
                'N7MR',
                'NBAC',
                'NBFT',
                'NBUS',
                'NCAS',
                'NCAT',
                'NCDI',
                'NCHE',
                'NCOW',
                'NCRN',
                'NDAV',
                'NDBR',
                'NDUK',
                'NELI',
                'NEWL',
                'NFAI',
                'NFBR',
                'NFIN',
                'NGRC',
                'NGRF',
                'NGUI',
                'NHIG',
                'NHOF',
                'NJCY',
                'NJES',
                'NLEX',
                'NMTI',
                'NNAC',
                'NNCP',
                'NNWB',
                'NPOC',
                'NRAV',
                'NRCK',
                'NREN',
                'NRUT',
                'NSND',
                'NSUN',
                'NTUR',
                'NTUS',
                'NTYL',
                'NUWH',
                'NWAY',
                'NWHI',
                'OXFO',
                'PLYM',
                'REED',
                'REID',
                'ROCK',
                'SALI',
                'SANP',
                'SCAR',
                'SCON',
                'SILR',
                'SKIN',
                'SPRU',
                'SSAT',
                'SSAV',
                'SSV1',
                'SSV2',
                'SWHI',
                'SWIT',
                'Station_Metadata.txt',
                'TAYL',
                'TBIG',
                'TBLE',
                'TBUR',
                'TCAM',
                'TCHC',
                'TCHE',
                'TCHU',
                'TCOK',
                'TCRO',
                'TFCL',
                'THAM',
                'TIND',
                'TLEW',
                'TMEI',
                'TMER',
                'TPRE',
                'TSHI',
                'TTOW',
                'TWES',
                'WAYN',
                'WHIT',
                'WILD',
                'WILL',
                'WINE'}
